// SPattern.cpp: implementation of the SPattern class.

#include "stdafx.h"

#define SPATTERN_C			// flag for global declarations

#include "convert.h"
#include "Geoutil.h"
#include "globals.h"
#include "Patterns.h"		// includes all pattern types
#include "SPos.h"
#include "Swathing.h"
#include "UnicodeUtil.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]=__FILE__;
	#define new DEBUG_NEW
#endif

//----------  Globals  ----------
//----------  Local Definitions  ----------

static const TCHAR s_szLLAfmt[] = _T("%lf %lf %f");

// Construction/Destruction

SPattern::SPattern()
{
	m_iMaxABLines		= 1;							// single pattern
	m_paPatInfo			= new PATINFO[m_iMaxABLines];	// allocate pattern info array
	
	m_csPatName.Empty();
	
	m_iPatId			= 0;
	m_ePatternType		= PATTYPE_UNKNOWN;		// unknown pattern type
	m_ePatDir			= PATDIR_INIT;			// unknown direction
	m_bClosedPattern	= false;
	m_iSwathNumBias		= 0;
	m_dOffsetM			= 0.0;
	m_iMaxSwathNo		= MAX_SWATH_NUM;

	m_iSavedSwathNum	= 0;
	m_iSavedSwathSeqIdx	= 0;

	m_dAreaSprayedHa	= 0.0;
	m_dAltitudeM		= 0.0;
	m_bIsJOBPat			= FALSE;
	m_bIsABCStepbackDone = FALSE;

	ClearPattern();								//	set defaults
}

SPattern::~SPattern()
{
	if ( m_paPatInfo != NULL )	{
		delete []m_paPatInfo;			// free the allocated pattern info array
		m_paPatInfo = NULL;
	}
}

//***********************************************************
/**	@brief	Get the number of active A-B lines

  This is a virtual method that can be overridden by
  derived classes.

  @return	int		= number of active A-B lines
*/
int SPattern::GetNumABLines()
{
	return m_iNumABLines;
}		// end GetNumABLines()

//***********************************************************
/**	@brief	Set the number of active A-B lines

  This is a virtual method that can be overridden by
  derived classes.

  @param	iNumLines		= new number of active A-B lines
*/
void SPattern::SetNumABLines( const int iNumLines )
{
	if ( (iNumLines >= 0)
	  && (iNumLines <= m_iMaxABLines) )
	{
		m_iNumABLines = iNumLines;
	}
}		// end SetNumABLines()


//***********************************************************
/**	@brief	Increment the number of active A-B lines
			in the current pattern.
*/
void SPattern::IncNumABLines()
{
	if ( m_iNumABLines < (m_iMaxABLines-1) )	{
		m_iNumABLines++;
	}
}		// end IncNumABLines()


//***********************************************************
/**	@brief	Get a pointer to the pattern info data

  This is a virtual method that can be overridden by 
  derived classes.

  @param	iPat		= array index of pattern info (normally 0 for single patterns)

  @returns	PPATINFO	= ptr to pattern info data block; NULL if error
*/
PPATINFO SPattern::GetPatternInfo( const int iPat )
{
	if ( (m_paPatInfo == NULL)
	  || (iPat < 0)
	  || (iPat >= m_iMaxABLines) )
	{
		return NULL;
	}

	return &(m_paPatInfo[iPat]);
}		// end GetPatternInfo()


//***********************************************************
/**	@brief	Clear the pattern

  @return	STATUS		= 0 if success; NZ if error
*/
STATUS SPattern::ClearPattern()
{
	STATUS	Stat = STAT_OK;

	m_iMaxSwathNo			= 0;
	m_iNumABLines			= 0;
	m_iCurrentSwathNum		= 0;
	m_iCurrentSwathSeqIdx	= 0;
	m_dSwathWidthM			= 65.0*FT2M;		// default swath width
	m_bIsJOBPat				= FALSE;
	m_bIsABCStepbackDone	= FALSE;

	ClearAllPatInfo();

	return Stat;
}		// end ClearPattern()


/************************************************************
/**	Clear all A-B line info
*/
void SPattern::ClearAllPatInfo()
{
	for ( int iPat=0; iPat < m_iMaxABLines; iPat++ )	{
		PPATINFO	ppat = GetPatternInfo( iPat );
		if ( ppat != NULL )		{
			ClearPatInfo( ppat );
		}		// end if ppat
	}
}		// end ClearAllPatInfo()


//-----------------------------------------------------------
/**	@brief	Clear a pattern info block

  @param	ppat		= ptr to PATINFO block
*/
void SPattern::ClearPatInfo( PPATINFO ppat )
{
	ppat->bAIsSet	= false;
	ppat->bBIsSet	= false;
	ppat->bCIsSet	= false;

	ppat->llaA.dLat = ppat->llaA.dLon = ppat->llaA.fAlt = 0.0;
	ppat->llaB.dLat = ppat->llaB.dLon = ppat->llaB.fAlt = 0.0;
	ppat->llaC.dLat = ppat->llaC.dLon = ppat->llaC.fAlt = 0.0;

	ppat->llaBkA.dLat = ppat->llaBkA.dLon = 0.0;		// Used for backup, OffSet IJMc
	ppat->llaBkB.dLat = ppat->llaBkB.dLon = 0.0;		// Ditto  		

	ppat->dSwathWidthM			= m_dSwathWidthM;		// default pattern swath width
	ppat->dSwathBearingD		= 0.0;

	ppat->lrectExtents.dEast	= 0.0;
	ppat->lrectExtents.dWest	= 0.0;
	ppat->lrectExtents.dNorth	= 0.0;
	ppat->lrectExtents.dSouth	= 0.0;
}		// end ClearPatInfo()


//***********************************************************
/**	Get the swath sequence index for a swath number

  @param	iSwathNum	= swath number ( <0 = left of AB, >0 = right of AB )

  @return	int			= swath sequence number (0-based);
							MAX_SWATH_NUM+1 if error
*/
int SPattern::GetSequenceIndex( const int iSwathNum )
{
	int		iSeqIdx;

	for ( iSeqIdx = 0; iSeqIdx <= MAX_SWATH_NUM; iSeqIdx++ )		{
		if ( GetSwathNumber(iSeqIdx) == iSwathNum )	{
			//	found it
			return iSeqIdx;
		}		// end if GetSwathNumber
	}		// end for iSeqIdx

	//	try the "back" side of the pattern
	for ( iSeqIdx = -1; iSeqIdx >= -MAX_SWATH_NUM; iSeqIdx-- )		{
		if ( GetSwathNumber(iSeqIdx) == iSwathNum )	{
			//	found it
			return iSeqIdx;
		}		// end if GetSwathNumber
	}		// end for iSeqIdx

	//	if we get here, we didn't find the swath number
	return (MAX_SWATH_NUM + 1);
}		// end GetSequenceIndex()


//***********************************************************
/**	@brief Set the swath width

  @param	dSwathWidthM	= swath width (meters)
*/
void SPattern::SetSwathWidth( const double dSwathWidthM )
{
	if ( (g_oSwathing.m_poPattern != NULL)
	 &&  (g_oSwathing.m_poPattern->m_ePatternType == PATTYPE_BACK_TO_BACK_HBOOM )
	 &&  (g_oSprayBoom.GetBoomModeAuto()))
	{
		double dSwath = g_oSwathing.GetBOOMSeccionsSwath();

		if (dSwath > 0.0)
		{
			m_dSwathWidthM = dSwath;
		}
		else
		{
			m_dSwathWidthM = dSwathWidthM;
		}
	}
	else
	{
		m_dSwathWidthM = dSwathWidthM;
	}

	// Set the field swath widths
	for ( int iField=0; iField < m_iNumABLines; iField++ )	{
		PPATINFO ppat = GetPatternInfo( iField );
		if ( ppat != NULL )	{
			ppat->dSwathWidthM = m_dSwathWidthM;	// default A-B swath width
		} // End if ppat
	}	// End for iField

}	// End SetSwathWidth()


//***********************************************************
/**	@brief	Get the swath width

  @return	double		= pattern swath width (meters)
*/
double SPattern::GetSwathWidth()
{
	return m_dSwathWidthM;
}		// end GetSwathWidth()


//***********************************************************
/*!	Calculate the physical swath from the sequence index

  \param	iSwathIdx	= swath sequence # in pattern (starting at 0)

  \return	int			= swath index ( <0 = left of AB, >0 = right of AB )
*/
int SPattern::GetSwathNumber( const int iSwathIdx)
	{
	if ( m_ePatDir == PATDIR_LEFT )	// pattern is to left
		return -iSwathIdx;
	else
		return iSwathIdx;
	}		// end GetSwathNumber()


// ***********************************************************
/*!	Get the range of the physical swath numbers in the current pattern

  negative swath numbers are to the left of the A-B line,
  positive swath numbers are to the right.

  This routine returns the numerical range, not the beginning and end of the sequence.
	
  \param	piMinSwath	= lowest swath number
  \param	piMaxSwath	= highest swath number

  \return	int			= 0 if success; NZ if error or no pattern
*/
int SPattern::GetSwathNumberRange( int* piMinSwath, int* piMaxSwath )
	{
	int		iStat = 0;

	if ( m_ePatDir == PATDIR_LEFT )		{
		m_iMaxSwathNo = -MAX_SWATH_NUM;
		*piMinSwath = -MAX_SWATH_NUM;
		*piMaxSwath = 0;
		}
	else	{
		m_iMaxSwathNo = MAX_SWATH_NUM;
		*piMinSwath = 0;
		*piMaxSwath = MAX_SWATH_NUM;
		}

	return iStat;
	}		// end GetSwathNumberRange()


//************************************************************************
/*!	Check if swath is in or out of the pattern

  \param	iSwathIdx	= pattern sequence #

  \return	bool		= T/F: in pattern
*/
bool SPattern::IsInPattern( const int iSwathIdx )
	{
	return ( iSwathIdx >= 0 );
	}		// end IsInPattern()


//************************************************************************
/*!	Check if swath is the last swath of the pattern

  This is a virtual method that may be overridden by derived classes.

  \param	iSwathIdx	= pattern sequence #

  \return	bool		= T/F: last swath
*/
bool SPattern::IsLastSwath( const int iSwathIdx )
	{
	//	default: see if the next swath in sequence will be outside the pattern
	return ( (IsClosedPattern())
		  && (IsInPattern(iSwathIdx))
		  && (! IsInPattern(iSwathIdx+1)) );
	}		// end IsLastSwath()


//******************************************************************
/**	Get array index from pattern type

  This is a static method.

  @param	ePatType		= pattern type enum

  @return	int				= log ID value
*/
int SPattern::PatType2Id( const PATTYPE ePatType )
{
	int	idx = (int)ePatType;

	if ( (idx >= 0)
	  && (idx < NUM_PATTYPES) )
	{
		return g_aiPatternTypeId[idx];
	}

	return 0;				// default
}		// end PatType2Id()


//******************************************************************
/**	Get pattern enum from log ID value

  This is a static method.

  @param	iPatId		= log ID value

  @return	PATTYPE		= pattern type enum
*/
PATTYPE SPattern::Id2PatType( const int iPatId )
{
	PATTYPE ePatType = PATTYPE_UNKNOWN;		// default

	//	scan to see which entry this is
	for ( int idx=0; idx < NUM_PATTYPES; idx++ )	{
		if ( g_aiPatternTypeId[idx] == iPatId )	{
			//	found it
			ePatType = (PATTYPE)idx;
			break;
		}		// end if g_aePatterns
	}		// end for idx

	return ePatType;
}		// end Index2PatType()


//**********************************************************
/**	Get Pattern Type from abbreviation.

  This is a static method.

  @param	pstr		= ptr to type abbreviation string

  @return	PATTYPE		= pattern type 
*/
PATTYPE SPattern::Name2PatType( LPCTSTR pstr )
{
	PATTYPE ePatType = PATTYPE_UNKNOWN;		// default
	CString csType = pstr;

	csType.TrimLeft();						// remove leading whitespace
	csType.TrimRight();						// remove trailing whitespace

	//	convert pre-2.1.5 AirTrac pattern type mnemonics to AirStar-compatable ones
	if      ( csType == "BACK-TO-BACK" )		csType = "BACK_TO_BACK";
	else if ( csType == "QUICK" )				csType = "QUICK_RACETRACK";
	else if ( csType == "REVERSE_RACETRACK" )	csType = "REV_RACETRACK";

	//	scan to see which entry this is
	//	must be an exact match (to distinguish things like 
	//	BACK_TO_BACK vs. BACK-TO-BACK-SKIP)
	for ( int idx=0; idx < NUM_PATTYPES; idx++ )	{
		//	ignore any extra characters after the type mnemonic
		if ( _tcsicmp(g_aszPatternTypeName[idx],csType) == 0 )	{
			//	found it
			ePatType = (PATTYPE)idx;
			break;
		}		// end if g_aePatterns
	}		// end for idx

	return ePatType;
}		// end Name2PatType()


//******************************************************************
/**	Get abbreviation from pattern type

  This is a static method.

  @param	ePatType		= pattern type enum

  @return	LPCTSTR			= ptr to abbreviation string
*/
LPCTSTR SPattern::PatType2Name( const PATTYPE ePatType )
{
	int	idx = (int)ePatType;

	if ( (idx >= 0)
	  && (idx < NUM_PATTYPES) )
	{
		return g_aszPatternTypeName[idx];
	}

	return g_aszPatternTypeName[(int)PATTYPE_UNKNOWN];		// default
}		// end PatType2Name()

//******************************************************************
/**	Get abbreviation from pattern type

  This is a static method.
  @param	ePatType		= pattern type enum
  @return	LPCTSTR			= ptr to abbreviation string
*/
LPCTSTR SPattern::PatType2Abbrev( const PATTYPE ePatType )
{
	int	idx = (int)ePatType;

	if ( (idx >= 0)
	  && (idx < NUM_PATTYPES) )
	{
		return g_aszPatternTypeAbbrev[idx];
	}

	return g_aszPatternTypeAbbrev[(int)PATTYPE_UNKNOWN];		// default

}		// end PatType2Abbrev()

/*---------------------------------------------------------------------------------------
	Abbrev2PatType() Get pattern from abbreviation									IJMc
	@param	LPCTSTR			= ptr to abbreviation string
	@return	PATTYPE			= pattern type
*///-------------------------------------------------------------------------------------
PATTYPE SPattern :: Abbrev2PatType( LPCTSTR pstr )
{
	PATTYPE ePatType = PATTYPE_UNKNOWN;		// default
	CString csType = pstr;

	csType.TrimLeft();			
	csType.TrimRight();

	int i = 0;

	for (i = 0; i < NUM_PATTYPES; i++ )	{
		if ( _tcsicmp(g_aszPatternTypeAbbrev[i],csType) == 0 )	{
			ePatType = (PATTYPE)i;
			break;
		}	// End if
	}	// End for i

	return(ePatType);

}	// End Abbrev2PatType()

//********************************************************************
/**	@brief	Set a pattern mark

  Default method: patterns with one A-B line.

  Virtual method: may be overloaded in derived classes to write 
  additional settings.

  @param	iMarkIdx	= pattern mark index: 0=A, 1=B, 2=C.
  @param	lla			= mark lat/lon/alt

  @return	STATUS		= 0 if success; NZ if error
*/
STATUS SPattern::SetMark( const int iMarkIdx, const LOC3& lla )
{
	STATUS	Stat = STAT_OK;

	int iField = 0;					// just the first field
	PPATINFO ppat = GetPatternInfo( iField );
	if ( ppat == NULL )		{
		return STAT_PATTERN_NO_AB_LINE;
	}

	switch ( iMarkIdx )		{
		//------------------
		case 0:						// "A"		
			ppat->llaA = lla;	
			ppat->llaBkA = lla;		// Original "A" coord  

			ppat->bAIsSet = true;	
		
			//	the rest of the points are not set up yet
			ppat->bBIsSet = false;
			ppat->bCIsSet = false;
			break;
			
		//------------------
		case 1:						// "B"	
			ppat->llaB = lla;	
			ppat->llaBkB = lla;		// Original "B" coord  
			ppat->bBIsSet = true;

			ppat->dSwathWidthM = m_dSwathWidthM;	// init pattern swath width

			if ( ppat->bAIsSet ) {
				ppat->dSwathBearingD = GeoutilCalculateBearing( ppat->llaA, ppat->llaB ); // A -> B bearing

				// Check to see if SetMark was called from a MARK/RETURN rutine
				if (!g_oMarks.GetIsMarkActive()) {
					g_oSwathing.OFFSetABPoints(this);	// Do A/B line OffSet IJMc
				}

				//	we have at least one A-B line now.
				if ( m_iNumABLines < 1 )	{
					m_iNumABLines = 1;
				}
			}		// end if bAIsSet
			break;
			
		//------------------
		case 2:					// "C"	
			ppat->llaC = lla;	
			ppat->bCIsSet = true;	
			Stat = SetUpClosedPattern( iField );
			break;

		//------------------
		default:
			Stat = STAT_NOTOK;
			break;
	}		// end switch iMarkIdx

	CalculatePatternExtents( ppat );	
	
	return Stat;
}		// end SetMark()


//--------------------------------------------------------------------
/**	Set up Closed Pattern

  @pre	The specified field has been initialized with the A, B and C points.

  @param	iField		= Field (A-B line) index number

  @return	STATUS		= 0 if success; NZ if error
*/
STATUS SPattern::SetUpClosedPattern( const int iField )
{
	PPATINFO pFld = GetPatternInfo( iField );
	if ( pFld == NULL )		{
		return STAT_PATTERN_NO_AB_LINE;
	}

	double dOffsetM	= CalculateOffsetFromSwath( iField, 0, pFld->llaC );	// offset from A-B line
	double dSwathWidthM = pFld->dSwathWidthM;		// swath width in meters
	
	if ( dSwathWidthM < 1.0 )	{
		// default to avoid divide by zero
		if ( m_dSwathWidthM > 1.0 )		{
			dSwathWidthM		= m_dSwathWidthM;		
		}	
		else	{
			dSwathWidthM		= 60.0*FT2M;
		}

		pFld->dSwathWidthM	= dSwathWidthM;
	}

	// Convert swath sequence index to swath number (0-based)
	short iSwath;
	double dLines = (dOffsetM / dSwathWidthM);
	double dWhole;
	double dFac = fabs(modf(dLines, &dWhole));
	
	if (dFac > 0.0 && dFac < 0.5 && g_oSwathing.m_bIsExtendC)
		iSwath = (short) (fabs(dWhole) +1);
	else
		iSwath = (short)floor( fabs(dOffsetM)/dSwathWidthM + 0.5 );

	if ( dOffsetM < 0.0 )
		iSwath = (short)( -iSwath );		// Make it negative again
	
	m_iMaxSwathNo = iSwath;					// This is the max swath number
	
	// Set left/right based on 'C' point
	if (iSwath == 0)
		m_ePatDir = PATDIR_INIT;
	else if (iSwath < 0)
		m_ePatDir = PATDIR_LEFT;
	else if (iSwath > 0)
		m_ePatDir = PATDIR_RIGHT;

	return STAT_OK;

}	// End SetUpClosedPattern()


//********************************************************************
/*!			Load the pattern from a file

  Virtual method: may be overloaded in derived classes to write 
  additional settings.

  @pre	Assumes that f is positioned to (re)read the .PAT line;
		if not, the caller has the responsibility to set up
		the pattern ID and label, if needed.

  @post	Leaves file ready to read the next line after the pattern data.

  @param	f			= CStdioFile object

  @return	STATUS		= 0 if success; NZ if error
*/
STATUS SPattern::LoadFromFile( HFile& f )
{
	#undef DBGLBL
	#define DBGLBL	"SPattern::LoadFromFile"

	int		idx;
	CString	csRec;
	CString	csData;

	m_iSavedSwathNum		= 0;				// default
	m_iSavedSwathSeqIdx		= 0;				// default
	
	if ( f.ReadLine(csRec) )		{

		TRACE1( "\t<< \"%s\"\n", csRec );

		csRec.TrimLeft();			// skip leading whitespace

		if ( csRec.Left(4) == _T(".PAT") )		{
			//	pattern header record
			idx = csRec.Find (' ',4);	// Find the first space
			csData = csRec.Mid(idx+1);	// extract the data portion of the record
			csData.TrimLeft();

			int		iPatID = 0;
			TCHAR	szName[80] = _T("");
		
			_stscanf_s( (LPCTSTR)csData, _T("%d %s"), &iPatID, szName );

			m_iPatId		= iPatID;
			m_csPatName		= szName;
		}		// end if .PAT
		else	{
			//	back up to re-read the line
			f.UndoReadLine();
		}
	}		// end if f.ReadLine

	//	the next line should be the pattern type
	f.ReadLine( csRec );
	TRACE1( "\t<< \"%s\"\n", csRec );
	//	<< we had to know the pattern type when we created
	//		this object, so ignore this line now.

	//	the next line should be the direction
	if ( f.ReadLine(csRec) )		{
		TRACE1( "\t<< \"%s\"\n", csRec );

		csRec.TrimLeft();			// skip leading whitespace
		csData = csRec.Left(4);
		csData.MakeUpper();

		if ( csData == _T("LEFT") )			{
			m_ePatDir = PATDIR_LEFT;
		}
		else if ( csData == _T("RIGH") )	{ // Only reading 4 chars
			m_ePatDir = PATDIR_RIGHT;
		}
		else if ( csData == _T("INIT") )	{
			m_ePatDir = PATDIR_RIGHT; // Default
		}
		else	{
			f.UndoReadLine();				// back up and try the line again
		}
	}

	//	read until we hit the start of the next entity
	while ( f.ReadLine( csRec ) )	{		// removes the CR/LF
		csRec.TrimLeft();			// skip leading whitespace
		if ( csRec.IsEmpty() )		{
			continue;				// blank line - ignore it
		}

		if ( csRec[0] == '.' )		{
			//	start of next file entity
			f.UndoReadLine();	//	back up to re-read the line
			break;
		}		// end if csRec[0]

		if ( ProcessFileLine(csRec) != STAT_OK )	{
			//	end of data or error
			break;
		}
	}		// end while f.ReadLine

	SetUpPattern();						//	calculate pattern parameters

	//	set the current swath
	//	(may have been reset to 0 by common code)
	m_iCurrentSwathNum		= m_iSavedSwathNum;
	m_iCurrentSwathSeqIdx	= m_iSavedSwathSeqIdx;

	return STAT_OK;

}	// End LoadFromFile()


//=================================================================
/**	@brief	Process a Pattern input line

  Virtual method: may be overloaded in derived classes to write 
  additional settings.

  @param	csRec		= input line

  @return	STATUS		= 0 if processed; NZ if end of data or not processed
*/
STATUS SPattern::ProcessFileLine( CString& csRec )
{
	STATUS	Stat = STAT_PATTERN_NOT_PROCESSED;
	CString	csKey;
	CString csKey3;
	CString csData;
	
	csKey.Empty();
	
	csRec.TrimLeft();
	
	int iKeyLen = csRec.Find(' ');		// find the first space, if any
	
	csKey = csRec.Left(iKeyLen);		// Get the entire key string
	csKey3 = csKey.Left(3);				// Get the first three characters		
	
	csData = csRec.Mid(iKeyLen+1);		// extract the data portion of the record
	csData.TrimLeft();					// trim any extra spaces at the start of the data
	
	//------------------------------
	if ( csKey == "SWIDTH" )	{		// SWIDTH = Swath Width
		if ( ! csData.IsEmpty() )	{
			double dWork = _tstof(csData);
			
			if ( csData.Find('M') >= 0 ) {// || csData.Find('m') >= 0)	{
				//	meters
				// <do nothing>
			}
			else if ( csData.Find('f') >= 0 || csData.Find('F') >= 0 )	{
				//	feet
				dWork *= FT2M;			// convert to meters
			}
			else	{
				//	assume feet
				dWork *= FT2M;			// convert to meters
			}

		SetSwathWidth( dWork );
		}		// end if NOT IsEmpty
		
		Stat = STAT_OK;						// done with the line
	}		// end if SWI
	
	/*****
	//------------------------------
	else if ( csKey3.Compare("MAN") == 0 )	{
		//	MANUAL_SWATH_NUMBER, PATTERN_SWATH_NUMBER, MAX_SWATH_NUMBER
		int		idx;
		
		if ( ! csData.IsEmpty() )	{
			m_iCurrentSwathNum	= atoi(csData);		// manual swath number
			m_iSavedSwathNum	= m_iCurrentSwathNum;
			
			idx = csData.Find( "PAT" );				// find the PATTERN_SWATH_NUMBER label
			if ( idx >= 0 )	{
				idx = csData.Find (' ',idx+3);		//Find the next space
				if (idx >= 0)		{
					m_iCurrentSwathSeqIdx	= atoi(csData.Mid(idx));	// manual swath index
					m_iSavedSwathSeqIdx		= m_iCurrentSwathSeqIdx;
				}
			}		// end if idx
			
			idx = csData.Find( "MAX",idx+1 );		// find the MAX_SWATH_NUMBER label
			if ( idx >= 0 )	{
				idx = csData.Find (' ',idx+3);		//Find the next space
				if (idx >= 0)		{
					m_iMaxSwathNo = atoi(csData.Mid(idx+1));	// max swath number
				}
			}		// end if idx
		}		// end if NOT IsEmpty
		
		TRACE3( "\t\tMSwNum = %d; PSwNum = %d; MaxSwNum = %d\n", 
			m_iCurrentSwathNum, m_iCurrentSwathSeqIdx, m_iMaxSwathNo );
		
		Stat = STAT_OK;						// done with the line
	}
	*****/

	//------------------------------
	else if ( csKey.Compare(_T("SNUMBER")) == 0 )	{
		//	swath number
		if ( ! csData.IsEmpty() )	{
			m_iCurrentSwathNum	= (short) _tstoi(csData);		// manual swath number
			m_iSavedSwathNum	= m_iCurrentSwathNum;	// save for end processing
		}
		
		Stat = STAT_OK;						// done with the line
	}

	//------------------------------
	else if ( csKey.Compare(_T("SINDEX")) == 0 )	{
		//	swath sequence index
		if ( ! csData.IsEmpty() )	{
			m_iCurrentSwathSeqIdx	= (short) _tstoi(csData);		// manual swath index
			m_iSavedSwathSeqIdx		= m_iCurrentSwathSeqIdx;	// save for end processing
		}
		
		Stat = STAT_OK;						// done with the line
	}

	//------------------------------
	else if ( csKey.Compare(_T("SMAX")) == 0 )	{
		//	max swath number
		if ( ! csData.IsEmpty() )	{
			m_iMaxSwathNo = (short) _tstoi(csData);	// max swath number
		}
		
		Stat = STAT_OK;						// done with the line
	}

	//------------------------------
	else if ( csKey.Compare(_T("SNBIAS")) == 0 )	{
		//	swath number bias
		m_iSwathNumBias = (short) _tstoi(csData);
		
		Stat = STAT_OK;						// done with the line
	}
	
	//------------------------------
	else if ( csKey3 == _T("ARE") )	{			// AREA_SPRAYED	= total area sprayed
		double dWork;
		
		if ( ! csData.IsEmpty() )	{
			dWork = _tstof(csData);
			
			//	check Hectares first, because it might have an 'A'
			if ( csData.Find('H') >= 0 )	{
				//	hectares
				// <do nothing>
			}
			else if ( csData.Find('A') >= 0 )	{
				//	acres
				dWork *= ACRES2HA;			// convert to hectares
			}
			else	{
				//	assume acres
				dWork *= ACRES2HA;			// convert to hectares
			}
			
			m_dAreaSprayedHa = dWork;
		}		// end if NOT IsEmpty
		
		Stat = STAT_OK;						// done with the line
	}		// end if ARE
	
	//------------------------------
	else if ( csKey3 == _T("FIE") )	{			// FIELD_WIDTH
		double dWork;
		
		if ( ! csData.IsEmpty() )	{
			dWork = _tstof(csData);
			
			// << ignore this for now >>
			
		}		// end if NOT IsEmpty
		
		Stat = STAT_OK;						// done with the line
	}		// end if FIE
	else if ( csKey == _T("OFFSET") )	{		// OFFSET  - swath line offset
		if ( ! csData.IsEmpty() )	{
			double dWork = _tstof(csData);
			
			if ( csData.Find('M') >= 0 || csData.Find('m') >= 0)	{
				//	meters 	// <do nothing>
			}
			else if ( csData.Find('f') >= 0 || csData.Find('F') >= 0 )	{
										// Feet
				dWork *= FT2M;			// Convert to meters
			}
			else	{ 					// Default assume feet
				dWork *= FT2M;			// Convert to meters
			}
			m_dOffsetM = dWork;
		}	// End if NOT IsEmpty
		Stat = STAT_OK;						// done with the line
	}	// End if OFFSET
	//------------------------------
	else if ( csKey.Compare(_T("ABLINES")) == 0 )	{	// ABLINES - number of A-B lines
		if ( ! csData.IsEmpty() )	{
			int	iWork = _tstoi(csData);
			if ( (iWork >= 0)
				&& (iWork <= m_iMaxABLines) )
			{
				m_iNumABLines = iWork;
			}
			else	{
				m_iNumABLines = m_iMaxABLines;		// limit to max # of lines
			}
			
		}		// end if NOT IsEmpty
		
		Stat = STAT_OK;						// done with the line
	}
	
	//------------------------------
	if ( csKey == _T("ALTITUDE") )	{			// ALTITUDE
		if ( ! csData.IsEmpty() )	{
			m_dAltitudeM = _tstof(csData);
			
			if ( csData.Find('M') >= 0 )	{
				//	meters
				// <do nothing>
			}
			else if ( csData.Find('F') >= 0 )	{
				//	feet
				m_dAltitudeM *= FT2M;			// convert to meters
			}
			else	{
				//	assume feet
				m_dAltitudeM *= FT2M;			// convert to meters
			}
		}		// end if NOT IsEmpty
		
		Stat = STAT_OK;						// done with the line
	}		// end if SWI
	
	//------------------------------
	else if ( iKeyLen == 1 )	{
		//	default pattern type - one A-B line
		//	A, B, C, ...
		int  iMarkIdx = (int)(csKey[0] - 'A');

		if ( (iMarkIdx >= 0)
			/*&& (iMarkIdx < SMARK_MAX_LOCS)*/ )
		{
			if ( ! csData.IsEmpty() )	{
				SPos	spos;

				spos.fAlt	= 0.0;			// default altitude
				
				int iNumFlds = _stscanf_s( (LPCTSTR)csData, s_szLLAfmt, 
					&spos.dLat, &spos.dLon, &spos.fAlt );
				
				if ( iNumFlds > 1 )		{
					// we have at least lat and lon
					if ( spos.IsValid2D() )	{			//	the location is valid
						SetMark( iMarkIdx, spos );
					}		// end if IsValid2D()
				}		// end if iNumFlds
				else	{
					//	LTRACE2( "Only %d fields found in \"%s\"\n", iNumFlds, csData );
				}
			}		// end if idx
		}		// end if iMarkIdx
		
		Stat = STAT_OK;						// done with the line
	}		// end else if iKeyLen == 1
	
	return Stat;
}		// end ProcessFileLine()
	

//*****************************************************************
/*!			Save the pattern to a file

  Virtual method: may be overloaded in derived classes to write 
  additional settings.

  @pre	Assumes that HFile f is open and positioned where
		the pattern block is to be written.

  @param	f				= HFile object
  @param	bWriteHeader	= t/f: write the .PAT header line

  @return	STATUS			= 0 if success; NZ if error
*/
STATUS SPattern::SaveToFile( HFile& f, bool bWriteHeader /*= true*/ )
{
	#undef DBGLBL
	#define DBGLBL	"SPattern::SaveToFile"

	CString csLine;

	if ( bWriteHeader )		{
		f.WriteLine( _T("\n") );			// put a blank line before the start of the pattern

		csLine.Format( _T(".PAT %d %s\n"), m_iPatId, m_csPatName );
		f.WriteLine( csLine );
	}		// end if bWriteHeader

	//------------------------------
	//	Pattern type
	csLine.Format( _T("   %s\n"), PatType2Name(m_ePatternType) );
	f.WriteString(csLine);

	//------------------------------
	//	Left/Right
	CString csSide;
	
	csSide  = g_apszPatDirNames[m_ePatDir];
	// This happens if [Pattern] Pattern_Direction=INIT
	// Must onlt save LEFT or RIGHT
	if ( (csSide != _T("LEFT"))
	&&   (csSide != _T("RIGHT"))) {
		csSide = _T("RIGHT");
	}
	csLine.Format( _T("   %s\n"), csSide);
	f.WriteString(csLine);

	//------------------------------
	//	Swath width
	csLine.Format( _T("   SWIDTH %.4lf M\n"), GetSwathWidth() );
	f.WriteString(csLine);

	//------------------------------
	//	Field width
	/***	ignore this for now
	csLine.Format( "   FIELD_WIDTH %.3lf M\n", m_dSwathWidthM );
	f.WriteString(csLine);
	***/

	//------------------------------
	//	Offset
	csLine.Format( _T("   OFFSET %.4lf M\n"), m_dOffsetM );
	f.WriteString(csLine);

	/****
	//------------------------------
	//	Swath numbers
	csLine.Format( "   MANUAL_SWATH_NUMBER  %d  PATTERN_SWATH_NUMBER  %d MAX_SWATH_NUMBER %d\n", 
					m_iCurrentSwathNum, m_iCurrentSwathSeqIdx, m_iMaxSwathNo);  //Save Mark No. Example : Mar 11
	f.WriteString(csLine);
	****/

	//------------------------------
	//	Swath Number
	csLine.Format( _T("   SNUMBER %d\n"), m_iCurrentSwathNum );
	f.WriteString(csLine);

	//------------------------------
	//	Swath Sequence Index
	csLine.Format( _T("   SINDEX %d\n"), m_iCurrentSwathSeqIdx );
	f.WriteString(csLine);

	//------------------------------
	//	Max Swath Number
	csLine.Format( _T("   SMAX %d\n"), m_iMaxSwathNo );
	f.WriteString(csLine);

	//------------------------------
	//	Swath number accumulation bias
	csLine.Format( _T("   SNBIAS %d\n"), m_iSwathNumBias );
	f.WriteString(csLine);

	//------------------------------
	//	Altitude
	csLine.Format( _T("   ALTITUDE %.4lf M\n"), m_dAltitudeM );
	f.WriteString(csLine);

	//------------------------------
	//	Area Sprayed
	csLine.Format( _T("   AREA_SPRAYED %.4lf Ha\n"), m_dAreaSprayedHa );
	f.WriteString(csLine);

	//------------------------------
	//	Number of A-B lines
	csLine.Format( _T("   ABLINES %d\n"), m_iNumABLines );
	f.WriteString(csLine);

	//------------------------------
	//	Pattern points
	static TCHAR s_szSwathPointFmt[] = _T("   %c  %.7lf  %.7lf  %.3f\n");

	for ( int iPat=0; iPat < m_iNumABLines; iPat++ )	{
		TCHAR	cPt = (TCHAR)( 'A' + (iPat*2) );

		PPATINFO	ppat = GetPatternInfo( iPat );

		if ( ppat != NULL )		{

			if ( (g_AGDispSettings.m_eAGDMode != AGD_OFF)
			  && (theApp.m_pDispModel->GetStopStartADGModel()))
			{
				LOC LocOut;

				if ( ppat->bAIsSet )	
				{
					g_pView->CalcADG_OFFSetReverse(ppat->llaA.dLon, ppat->llaA.dLat, 
												   &LocOut.dLon, &LocOut.dLat);
					csLine.Format(s_szSwathPointFmt, cPt, LocOut.dLat, LocOut.dLon, ppat->llaA.fAlt);
					f.WriteString(csLine);
				}	// End bAIsSet

				if ( ppat->bBIsSet )	
				{
					g_pView->CalcADG_OFFSetReverse(ppat->llaB.dLon, ppat->llaB.dLat, 
												   &LocOut.dLon, &LocOut.dLat);
					csLine.Format(s_szSwathPointFmt, cPt+1, LocOut.dLat, LocOut.dLon, ppat->llaB.fAlt);
					f.WriteString(csLine);
				}	// End bBIsSet

				//	Currently, only single-field patterns can have 'C' points.
				if ( m_iNumABLines == 1 )	
				{
					if ( ppat->bCIsSet )	{
						g_pView->CalcADG_OFFSetReverse(ppat->llaC.dLon, ppat->llaC.dLat, 
													   &LocOut.dLon, &LocOut.dLat);
						csLine.Format(s_szSwathPointFmt, cPt+2, LocOut.dLat, LocOut.dLon, ppat->llaC.fAlt);
						f.WriteString(csLine);
					}	// End if bCIsSet
				}	// End if m_iNumABLines
			}
			else
			{
				if ( ppat->bAIsSet )	{
					csLine.Format( s_szSwathPointFmt, cPt, 
											ppat->llaA.dLat, ppat->llaA.dLon, ppat->llaA.fAlt );
					f.WriteString(csLine);
				}	// End bAIsSet

				if ( ppat->bBIsSet )	{
					csLine.Format( s_szSwathPointFmt, cPt+1, 
											ppat->llaB.dLat, ppat->llaB.dLon, ppat->llaB.fAlt );
					f.WriteString(csLine);
				}	// End bBIsSet

				//	Currently, only single-field patterns can have 'C' points.
				if ( m_iNumABLines == 1 )	{
					if ( ppat->bCIsSet )	{
						csLine.Format( s_szSwathPointFmt, cPt+2, 
												ppat->llaC.dLat, ppat->llaC.dLon, ppat->llaC.fAlt );
						f.WriteString(csLine);
					}	// End if bCIsSet
				}	// End if m_iNumABLines
			}
		}	// End if ppat
	}	// End for iPat

	return STAT_OK;

}	// End SaveToFile()


/***
////////////////////////////////////////////////////////////
//	Operators

SPattern& SPattern::operator =( const SPattern & pat )
	{
	Copy( pat );					// copy the object data
	
	return *this;
	}
***/


//=======================================================================
/**	@brief	Copy the object data from another SPattern object

  Virtual method: may be overloaded in derived classes to write 
  additional settings.

*/
void SPattern::Copy( const SPattern& pat )
{
	if ( this != &pat )		{
		m_iPatId			= pat.m_iPatId;
		m_csPatName			= pat.m_csPatName;
		m_ePatternType		= pat.m_ePatternType;
		m_bClosedPattern	= pat.m_bClosedPattern;
		m_ePatDir			= pat.m_ePatDir;
		m_iMaxSwathNo		= pat.m_iMaxSwathNo;
		m_iCurrentSwathNum	= pat.m_iCurrentSwathNum;
		m_iCurrentSwathSeqIdx = pat.m_iCurrentSwathSeqIdx;
		m_iSwathNumBias		= pat.m_iSwathNumBias;
		m_dSwathWidthM		= pat.m_dSwathWidthM;
		m_dOffsetM			= pat.m_dOffsetM;
		m_iNumABLines		= pat.m_iNumABLines;
		m_bIsJOBPat			= pat.m_bIsJOBPat;			// IJMc

		//	copy the pattern info blocks
		if ( m_iMaxABLines != pat.m_iMaxABLines )	{
			//	this probably shouldn't happen, but if it does...
			//	reallocate the pattern info array
			if ( m_paPatInfo != NULL )	{
				delete []m_paPatInfo;
			}

			m_iMaxABLines	= pat.m_iMaxABLines;
			m_paPatInfo		= new PATINFO[m_iMaxABLines];	// allocate pattern info array
		}		// end if m_iMaxABLines

		if ( (m_paPatInfo != NULL)
		  && (pat.m_paPatInfo != NULL) )
		{
			memcpy( m_paPatInfo, pat.m_paPatInfo, m_iMaxABLines*sizeof(PATINFO) );
		}		// end if m_paPatInfo

		///	@todo	copy pattern marks, etc.
	}
	
}		// end Copy()


//***********************************************************************
/**	Create a new copy of a pattern object

  This is a static method.

  @param	poOriginalPat	= ptr to original SPattern object

  @return	SPattern*		= ptr to object clone; NULL if error
*/
SPattern* SPattern::CloneObject( SPattern const * const poOriginalPat )
{
	if ( poOriginalPat == NULL )	{
		return NULL;
	}

	PATTYPE	eType = poOriginalPat->m_ePatternType;
	SPattern* poNewPat = CreatePatternObject( eType );

	if ( poNewPat != NULL )		{
	//	*poNewPat = *poOriginalPat;		// copy the original object to the new one.
		poNewPat->Copy( *poOriginalPat );	// copy the original data to the new object
	}		// end if poNewPat

	return poNewPat;
}		// end CloneObject()


//***********************************************************************
/**	Create a new pattern object using file information

  This is a static method.

  @pre	Assumes that f is positioned to (re)read the .PAT line;

  @post	Leaves file ready to read the next line after the pattern data.
  @post	Call GetLastError() to get error status if required.

  @param	f			= HFile object

  @return	SPattern*	= pointer to created pattern object; NULL if error
*/
SPattern* SPattern::CreateNewObjectFromFile( HFile& f )
{
	#undef DBGLBL
	#define DBGLBL	"SPattern::CreateNewObjectFromFile"

	m_statLast = STAT_OK;

	CString csLine;

	//	Read the next line and see if it's the pattern header
	//	Skip empty lines
	do	{
		if ( ! f.ReadLine(csLine) )		{
			//	error or EOF reading the line
			return NULL;
		}

		csLine.TrimLeft();
	}	while ( csLine.IsEmpty() );

	int			iPatID = 0;
	CString		csName;
	csName.Empty();
	SPattern*	poPat = NULL;
	CString		csData;

	//	the line could start with ".PAT" or just "PAT"
	if ( csLine.Left(1) == _T('.') )	{
		//	strip the period from the front of the line
		csLine = csLine.Mid(1);
	}

	if ( _tcsnicmp(csLine,_T("PAT"),3) != 0 )	{
		//	this is not a PATtern line; 
		//	abort because we can't determine the pattern type
		m_statLast = STAT_PATTERN_UNKNOWN_TYPE;
		return NULL;
	}

	int iFirstSpaceIdx = csLine.Find(' ');				// find the first space, if any

	csData = csLine.Mid(iFirstSpaceIdx+1);				// extract the data portion of the record
	csData.TrimLeft();

	_stscanf_s( csData, _T("%d"), &iPatID );
	int iFind = csData.Find(_T(' '), 0);
	CString csStr = csData.Mid(iFind);
	csName = csStr;

	//	read the next line to get the type of pattern
	if ( f.ReadLine(csLine) )		{
		TRACE1( "\t<< \"%s\"\n", csLine );
		
		csLine.TrimLeft();								// skip leading whitespace
		csLine.TrimRight();								// truncating trailing whitespace

		//	we should have a pattern type mnemonic
		PATTYPE ePatType = SPattern::Name2PatType(csLine);
		if ( ePatType == PATTYPE_UNKNOWN )		{
			//	bad pattern type: abort
			f.UndoReadLine();							// back up to re-read the line
			m_statLast = STAT_PATTERN_UNKNOWN_TYPE;
			return NULL;
		}		// end if ePatType

		//	create a pattern object of the proper type
		poPat = SPattern::CreatePatternObject( ePatType );
		if ( poPat != NULL )		{
			//	success creating the pattern object
			poPat->m_iPatId		= iPatID;
			poPat->m_csPatName	= csName;
			poPat->m_bIsJOBPat = TRUE;					// PATtern was created from a JOB file

			f.UndoReadLine();							// back up to re-read the line

			m_statLast = poPat->LoadFromFile( f );		// load the pattern data
			if ( m_statLast == STAT_OK )	{
				//	success loading the pattern info
			}
			else	{
				//	error loading the pattern info
				delete poPat;
				poPat = NULL;
			}	// End if Stat
		}	// End if poPat
	}	// End if ReadLine

	return poPat;
}	// End CreateNewObjectFromFile()


//********************************************************************
/**	@brief	Create a pattern object using pattern type.

  Creates an object of the correct pattern class per the specified
  pattern type.  This is a static method.

  @param	ePatType		= pattern type to create

  @return	SPattern*		= ptr to the new pattern object; 
								NULL if error
*/
SPattern* SPattern::CreatePatternObject( const PATTYPE ePatType )
{
	SPattern* poPattern;

	m_statLast = STAT_OK;

	switch ( ePatType )		{
		case PATTYPE_BACK_TO_BACK:
			poPattern = new SPatternB2B;
			break;

		case PATTYPE_BACK_TO_BACK_HBOOM:
			poPattern = new SPatternB2BHalfBoom;
			break;

		case PATTYPE_RACETRACK:
			poPattern = new SPatternRacetrack;
			break;

		case PATTYPE_SQUEEZE:
			poPattern = new SPatternSqueeze;
			break;

		case PATTYPE_QUICK_RACETRACK:
			poPattern = new SPatternQkRacetrack;
			break;

		case PATTYPE_REVERSE_RACETRACK:
			poPattern = new SPatternRevRacetrack;
			break;

		case PATTYPE_EXPAND:
			poPattern = new SPatternExpand;
			break;

		case PATTYPE_BACK_TO_BACK_SKIP:
			poPattern = new SPatternB2BSkip;
			break;

		case PATTYPE_MULTI_BACK_TO_BACK:	
			poPattern = new SPatternMultiB2B;
			break;

		case PATTYPE_QUICKTRACK_X:
			poPattern = new SPatternQuicktrackX( SPatternQuicktrackX::m_iNumQtxSwaths );
			SPatternQuicktrackX::m_iNumPatternSwaths = SPatternQuicktrackX::m_iNumQtxSwaths;
			break;

		case PATTYPE_NEAREST_SWATH:
			poPattern = new SPatternNS;
			break;

		default:
			//	unknown pattern type
			poPattern = NULL;
			m_statLast = STAT_PATTERN_UNKNOWN_TYPE;
			break;
	}		// end switch m_ePatternType

	return poPattern;
}		// end CreatePatternObject()


//***********************************************************************
/** @brief		Relocate a pattern to a new A-B line

  @param	iField		= pattern index
  @param	locNewA		= new "A"
  @param	locNewB		= new "B"

  @return	STATUS		= 0 if success; NZ if error
*/
STATUS SPattern::RelocatePattern( const int iField,
								 const LOC3 locNewA, const LOC3 locNewB )
{
	PPATINFO pFld = GetPatternInfo( iField );
	if ( pFld == NULL )	{
		return STAT_NOTOK;
	}

	//	change the A, B, and C points of the field
	double dCOffsetLat	= pFld->llaC.dLat - pFld->llaA.dLat; 
	double dCOffsetLon	= pFld->llaC.dLon - pFld->llaA.dLon;
	float fCOffsetAlt	= pFld->llaC.fAlt - pFld->llaA.fAlt;

	memcpy( &pFld->llaA, &locNewA, sizeof(LOC3) );
	memcpy( &pFld->llaB, &locNewB, sizeof(LOC3) );

	if ( pFld->bCIsSet )	{
		//	calculate the new C point
		pFld->llaC.dLat	= pFld->llaA.dLat + dCOffsetLat;
		pFld->llaC.dLon	= pFld->llaA.dLon + dCOffsetLon;
		pFld->llaC.fAlt	= pFld->llaA.fAlt + fCOffsetAlt;
	}

	SetUpPattern();						//	calculate pattern parameters

	return STAT_OK;
}		// end RelocatePattern()


//*****************************************************************
/**	@brief Set up pattern parameters

  Does necessary calculations after loading pattern data
*/
void SPattern::SetUpPattern()
{
	int		iField;

	//	calculate the bearing from A to B
	for ( iField=0; iField < m_iNumABLines; iField++ )	{
		PPATINFO pFld = GetPatternInfo( iField );
		if ( pFld != NULL )	{
			pFld->dSwathWidthM   = m_dSwathWidthM;		// default A-B swath width
			pFld->dSwathBearingD = GeoutilCalculateBearing( pFld->llaA, pFld->llaB ); // A -> B bearing
		}		// end if ppat
	}		// end for iField

}		// end SetUpPattern()


//**************************************************************
/*!	Get extents of pattern A, B, and C points

  Sets slrExtents in pattern info block

  \param	ppat		= ptr to pattern info block
*/
void SPattern::CalculatePatternExtents( PPATINFO ppat )
{
	if ( ! ppat->bAIsSet )		{
		// clear the extents
		ppat->lrectExtents.dEast	= 0.0;
		ppat->lrectExtents.dWest	= 0.0;
		ppat->lrectExtents.dNorth	= 0.0;
		ppat->lrectExtents.dSouth	= 0.0;
		return;
	}
	
	ppat->lrectExtents.dWest = ppat->lrectExtents.dEast = ppat->llaA.dLon;
	ppat->lrectExtents.dSouth = ppat->lrectExtents.dNorth = ppat->llaA.dLat;
	
	if ( ! ppat->bBIsSet )		{
		return;
	}
	
	/*!	\todo	Get rectangular extents of field
				to current swath instead of just 'C'
	*/
	//	Calculate extents of rectangle +/- n swaths around current swath
	int		iSwaths		= 20;									// number of swaths on each side of current
	int		iCurSwath	= m_iCurrentSwathNum;					// 0-based, <0 = left, >0 = right
	double	dSwathWidthM = ppat->dSwathWidthM;					// swath width in meters
	if ( dSwathWidthM < 1.0 )		dSwathWidthM = DEFAULT_SWATHWIDTH;		// prevent division by zero
	
	double	dSwAngR		= ppat->dSwathBearingD*D2R;
	double	dLonFact	= cos( ppat->llaA.dLat*D2R );
	
	double	dSwathWidthD = dSwathWidthM*M2DEG;					// swath width in degrees
	double	dSwOffLatD	= -dSwathWidthD*sin(dSwAngR);			// degrees offset in latitude
	double	dSwOffLonD	= (dSwathWidthD*cos(dSwAngR))/dLonFact;	// degrees offset in longitude
	
	LOC		loc;
	
	loc.dLat = ppat->llaA.dLat + (iCurSwath + iSwaths)*dSwOffLatD;
	loc.dLon = ppat->llaA.dLon + (iCurSwath + iSwaths)*dSwOffLonD;
	GeoutilExpandRectToLoc( ppat->lrectExtents, loc );
	
	loc.dLat = ppat->llaB.dLat + (iCurSwath + iSwaths)*dSwOffLatD;
	loc.dLon = ppat->llaB.dLon + (iCurSwath + iSwaths)*dSwOffLonD;
	GeoutilExpandRectToLoc( ppat->lrectExtents, loc );
	
	loc.dLat = ppat->llaA.dLat + (iCurSwath - iSwaths)*dSwOffLatD;
	loc.dLon = ppat->llaA.dLon + (iCurSwath - iSwaths)*dSwOffLonD;
	GeoutilExpandRectToLoc( ppat->lrectExtents, loc );
	
	loc.dLat = ppat->llaB.dLat + (iCurSwath - iSwaths)*dSwOffLatD;
	loc.dLon = ppat->llaB.dLon + (iCurSwath - iSwaths)*dSwOffLonD;
	GeoutilExpandRectToLoc( ppat->lrectExtents, loc );
	
	//	Show the C point if it exists
	if ( ppat->bCIsSet )		{
		GeoutilExpandRectToLoc( ppat->lrectExtents, ppat->llaC );
	}
}		// end CalculatePatternExtents()
	
	
//**********************************************************************
/*!	Calculate Offset from swath

  @param	iField		= pattern index
  @param	iSwath		= physical swath (<0 = left of A-B, >0 = right)
  @param	loc			= lat/lon/alt to calculate
  
  @return	double		= crosstrack distance relative to A -> B direction (meters)
*/
double SPattern::CalculateOffsetFromSwath( const int iField, const int iSwath, const LOC3& loc )
{
	#undef DBGLBL
	#define DBGLBL "CAirTracView::CalculateOffsetFromSwath"
		
	if ( ! GeoutilValidLL(loc) )		{
		//-----  Position is invalid  -----
		return 0.0;
	}
	
	PATINFO*	ppat = GetPatternInfo( iField );
	if ( ppat == NULL )		{
		return 0.0;
	}

	double dLatR = loc.dLat*D2R;							// radians
	double dLonR = loc.dLon*D2R;							// radians
	double dAltM = loc.fAlt;								// meters
	
	double dSwathAngR	= ppat->dSwathBearingD*D2R;	// swath angle (radians)
	
	double dRpmM;
	
	//	We only need to calculate the rotation functions (sin and cos)
	//	when the swath angle changes (new pattern).
	
	static double s_dLastSwathAngR = 0.0;
	static double s_dSin = sin(0.0);
	static double s_dCos = cos(0.0);
	
	if ( dSwathAngR != s_dLastSwathAngR )	{
		//	the swath angle has changed; calculate the new parameters
		//	needed to rotate the swath orientation back to 0
		s_dSin = sin(-dSwathAngR);						// reverse angle
		s_dCos = cos(-dSwathAngR);
		s_dLastSwathAngR = dSwathAngR;
	}		// end if dSwathAngR
	
	GeoutilGeoidRadii( dLatR, dAltM, &dRpmM, NULL );		// get radius of prime meridian
	
	double	dLonFact	= cos(dLatR);
	
	double	dSwathWidthR	= ppat->dSwathWidthM*M2DEG*D2R;	// swath width in radians
	double	dSwOffLatR	= -dSwathWidthR*sin(dSwathAngR);
	double	dSwOffLonR	= (dSwathWidthR*cos(dSwathAngR))/dLonFact;
	
	//	calculate the A point of the current swath
	double dAPtLatR		= ppat->llaA.dLat*D2R + (double)iSwath*dSwOffLatR;
	double dAPtLonR		= ppat->llaA.dLon*D2R + (double)iSwath*dSwOffLonR;
	
	double	dLat0R		= dLatR - dAPtLatR;					// translate so that 'A' is origin
	double	dLon0R		= (dLonR - dAPtLonR)*dLonFact;		// correct for longitude convergence
	
	//	the corrected longitude difference after reverse rotation (in radians)
	//	will be the crosstrack relative to the A -> B line orientation.
	//	Multiply this by the radius of the prime meridian to get meters.
	double dOffsetM = ((s_dCos*dLon0R) + (s_dSin*dLat0R))*dRpmM;	// offset from swath
	
	#if DEBUG_TRACE_ENABLED && DBG_VERBOSE	//=======================================
		if ( s_iDbgPatCount == 0 )	{
			TRACE1( "offset %.6lf\n", dOffsetM );
		}
	#endif									//=======================================
	
	return dOffsetM;	// offset from swath: relative to A -> B direction
}		// end CalculateOffsetFromSwath()


//***********************************************************************
/**	@brief		Get the closest swath number to a location

  @param	iField		= pattern index
  @param	loc			= position

  @return	int			= physical swath (<0 = left of A-B, >0 = right)
*/
int SPattern::CalculateNearestSwath( const int iField, const LOC3& loc )
{
	double	dABXTrack = CalculateOffsetFromSwath( iField, 0, loc );
	int	iSwathNum = (int)floor( (dABXTrack/GetSwathWidth()) + 0.5 );

	//	Swath number can be negative (decrement past A-B)
	if      ( iSwathNum < -999 )	iSwathNum = -999;
	else if ( iSwathNum > 999 )		iSwathNum = 999;

	return iSwathNum;
}		// end CalculateNearestSwath()


//***********************************************************************
/** @brief		Calculate the "A" and "B" points of a swath

  @param	iField		= pattern index
  @param	iSwath		= physical swath ( <0 = left of A-B, >0 = right)
  @param	locA		= LOC3 struct to receive "A"
  @param	locB		= LOC3 struct to receive "B"

  @return	STATUS		= 0 if success; NZ if error
*/
STATUS SPattern::CalculateSwathEnds( const int iField, const int iSwath,
									LOC3& locA, LOC3& locB )
{
	STATUS Stat = STAT_OK;

	PATINFO*	ppat = GetPatternInfo( iField );
	if ( ppat == NULL )		{
		return STAT_NOTOK;
	}

	double	dSwathWidthM = ppat->dSwathWidthM;					// swath width in meters
	if ( dSwathWidthM < 1.0 )		dSwathWidthM = DEFAULT_SWATHWIDTH;		// prevent division by zero
	
	double	dSwAngR		= ppat->dSwathBearingD*D2R;
	double	dLonFact	= cos( ppat->llaA.dLat*D2R );
	
	double	dSwathWidthD = dSwathWidthM*M2DEG;					// swath width in degrees
	double	dSwOffLatD	= -dSwathWidthD*sin(dSwAngR);			// degrees offset in latitude
	double	dSwOffLonD	= (dSwathWidthD*cos(dSwAngR))/dLonFact;	// degrees offset in longitude
			
	//	calculate the swath endpoints
	if ( ppat->bAIsSet )	{
		locA.dLat	= ppat->llaA.dLat + ((double)iSwath*dSwOffLatD);
		locA.dLon	= ppat->llaA.dLon + ((double)iSwath*dSwOffLonD);
		locA.fAlt	= ppat->llaA.fAlt;
	}
	else	{
		//	the pattern is not set up
		locA.dLat	= 0.0;
		locA.dLon	= 0.0;
		locA.fAlt	= 0.0;

		Stat = STAT_PATTERN_NO_AB_LINE;
	}

	if ( ppat->bBIsSet )	{
		locB.dLat	= ppat->llaB.dLat + ((double)iSwath*dSwOffLatD);
		locB.dLon	= ppat->llaB.dLon + ((double)iSwath*dSwOffLonD);
		locB.fAlt	= ppat->llaB.fAlt;
	}
	else	{
		//	no B
		locB.dLat	= 0.0;
		locB.dLon	= 0.0;
		locB.fAlt	= 0.0;

		Stat = STAT_PATTERN_NO_B;
	}

	return Stat;
}		// end CalculateSwathEnds()

